/*
 * Copyright (c) 2011-2013, Peter Abeles. All Rights Reserved.
 *
 * This file is part of BoofCV (http://boofcv.org).
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package boofcv.alg.filter.convolve.border;

import boofcv.core.image.border.ImageBorder_F32;
import boofcv.core.image.border.ImageBorder_I32;
import boofcv.struct.convolve.Kernel1D_F32;
import boofcv.struct.convolve.Kernel1D_I32;
import boofcv.struct.convolve.Kernel2D_F32;
import boofcv.struct.convolve.Kernel2D_I32;
import boofcv.struct.image.ImageFloat32;
import boofcv.struct.image.ImageInt16;
import boofcv.struct.image.ImageSInt32;

/**
 * <p>
 * Convolves just the image's border.  How the border condition is handled is specified by the {@link boofcv.core.image.border.ImageBorder}
 * passed in.  For 1D kernels only the horizontal or vertical borders are processed.
 * </p>
 * 
 * <p>
 * WARNING: Do not modify.  Automatically generated by {@link GenerateConvolveJustBorder_General}.
 * </p>
 * 
 * @author Peter Abeles
 */
public class ConvolveJustBorder_General {

	public static void horizontal(Kernel1D_F32 kernel, ImageBorder_F32 input, ImageFloat32 output , int border ) {
		final float[] dataDst = output.data;
		final float[] dataKer = kernel.data;

		final int radius = kernel.getRadius();
		final int width = output.getWidth();
		final int height = output.getHeight();

		for (int y = 0; y < height; y++) {
			int indexDest = output.startIndex + y * output.stride;

			for ( int x = 0; x < border; x++ ) {
				float total = 0;
				for (int k = -radius; k <= radius; k++) {
					total += input.get(x+k,y) * dataKer[k+radius];
				}
				dataDst[indexDest++] = total;
			}

			indexDest = output.startIndex + y * output.stride + width-border;
			for ( int x = width-border; x < width; x++ ) {
				float total = 0;
				for (int k = -radius; k <= radius; k++) {
					total += input.get(x+k,y) * dataKer[k+radius];
				}
				dataDst[indexDest++] = total;
			}
		}
	}

	public static void vertical(Kernel1D_F32 kernel, ImageBorder_F32 input, ImageFloat32 output , int border ) {
		final float[] dataDst = output.data;
		final float[] dataKer = kernel.data;

		final int radius = kernel.getRadius();
		final int width = output.getWidth();
		final int height = output.getHeight();

		for ( int x = 0; x < width; x++ ) {
			int indexDest = output.startIndex + x;

			for (int y = 0; y < border; y++, indexDest += output.stride) {
				float total = 0;
				for (int k = -radius; k <= radius; k++) {
					total += input.get(x,y+k) * dataKer[k+radius];
				}
				dataDst[indexDest] = total;
			}

			indexDest = output.startIndex + (height-border) * output.stride + x;
			for (int y = height-border; y < height; y++, indexDest += output.stride) {
				float total = 0;
				for (int k = -radius; k <= radius; k++ ) {
					total += input.get(x,y+k) * dataKer[k+radius];
				}
				dataDst[indexDest] = total;
			}
		}
	}

	public static void convolve(Kernel2D_F32 kernel, ImageBorder_F32 input, ImageFloat32 output , int border ) {
		final float[] dataDst = output.data;
		final float[] dataKer = kernel.data;

		final int radius = kernel.getRadius();
		final int width = output.getWidth();
		final int height = output.getHeight();

		// convolve along the left and right borders
		for (int y = 0; y < height; y++) {
			int indexDest = output.startIndex + y * output.stride;

			for ( int x = 0; x < border; x++ ) {
				float total = 0;
				int indexKer = 0;
				for( int i = -radius; i <= radius; i++ ) {
					for (int j = -radius; j <= radius; j++) {
						total += input.get(x+j,y+i) * dataKer[indexKer++];
					}
				}
				dataDst[indexDest++] = total;
			}

			indexDest = output.startIndex + y * output.stride + width-border;
			for ( int x = width-border; x < width; x++ ) {
				float total = 0;
				int indexKer = 0;
				for( int i = -radius; i <= radius; i++ ) {
					for (int j = -radius; j <= radius; j++) {
						total += input.get(x+j,y+i) * dataKer[indexKer++];
					}
				}
				dataDst[indexDest++] = total;
			}
		}

		// convolve along the top and bottom borders
		for ( int x = border; x < width-border; x++ ) {
			int indexDest = output.startIndex + x;

			for (int y = 0; y < border; y++, indexDest += output.stride) {
				float total = 0;
				int indexKer = 0;
				for( int i = -radius; i <= radius; i++ ) {
					for (int j = -radius; j <= radius; j++) {
						total += input.get(x+j,y+i) * dataKer[indexKer++];
					}
				}
				dataDst[indexDest] = total;
			}

			indexDest = output.startIndex + (height-border) * output.stride + x;
			for (int y = height-border; y < height; y++, indexDest += output.stride) {
				float total = 0;
				int indexKer = 0;
				for( int i = -radius; i <= radius; i++ ) {
					for (int j = -radius; j <= radius; j++) {
						total += input.get(x+j,y+i) * dataKer[indexKer++];
					}
				}
				dataDst[indexDest] = total;
			}
		}
	}

	public static void horizontal(Kernel1D_I32 kernel, ImageBorder_I32 input, ImageInt16 output , int border ) {
		final short[] dataDst = output.data;
		final int[] dataKer = kernel.data;

		final int radius = kernel.getRadius();
		final int width = output.getWidth();
		final int height = output.getHeight();

		for (int y = 0; y < height; y++) {
			int indexDest = output.startIndex + y * output.stride;

			for ( int x = 0; x < border; x++ ) {
				int total = 0;
				for (int k = -radius; k <= radius; k++) {
					total += input.get(x+k,y) * dataKer[k+radius];
				}
				dataDst[indexDest++] = (short)total;
			}

			indexDest = output.startIndex + y * output.stride + width-border;
			for ( int x = width-border; x < width; x++ ) {
				int total = 0;
				for (int k = -radius; k <= radius; k++) {
					total += input.get(x+k,y) * dataKer[k+radius];
				}
				dataDst[indexDest++] = (short)total;
			}
		}
	}

	public static void vertical(Kernel1D_I32 kernel, ImageBorder_I32 input, ImageInt16 output , int border ) {
		final short[] dataDst = output.data;
		final int[] dataKer = kernel.data;

		final int radius = kernel.getRadius();
		final int width = output.getWidth();
		final int height = output.getHeight();

		for ( int x = 0; x < width; x++ ) {
			int indexDest = output.startIndex + x;

			for (int y = 0; y < border; y++, indexDest += output.stride) {
				int total = 0;
				for (int k = -radius; k <= radius; k++) {
					total += input.get(x,y+k) * dataKer[k+radius];
				}
				dataDst[indexDest] = (short)total;
			}

			indexDest = output.startIndex + (height-border) * output.stride + x;
			for (int y = height-border; y < height; y++, indexDest += output.stride) {
				int total = 0;
				for (int k = -radius; k <= radius; k++ ) {
					total += input.get(x,y+k) * dataKer[k+radius];
				}
				dataDst[indexDest] = (short)total;
			}
		}
	}

	public static void convolve(Kernel2D_I32 kernel, ImageBorder_I32 input, ImageInt16 output , int border ) {
		final short[] dataDst = output.data;
		final int[] dataKer = kernel.data;

		final int radius = kernel.getRadius();
		final int width = output.getWidth();
		final int height = output.getHeight();

		// convolve along the left and right borders
		for (int y = 0; y < height; y++) {
			int indexDest = output.startIndex + y * output.stride;

			for ( int x = 0; x < border; x++ ) {
				int total = 0;
				int indexKer = 0;
				for( int i = -radius; i <= radius; i++ ) {
					for (int j = -radius; j <= radius; j++) {
						total += input.get(x+j,y+i) * dataKer[indexKer++];
					}
				}
				dataDst[indexDest++] = (short)total;
			}

			indexDest = output.startIndex + y * output.stride + width-border;
			for ( int x = width-border; x < width; x++ ) {
				int total = 0;
				int indexKer = 0;
				for( int i = -radius; i <= radius; i++ ) {
					for (int j = -radius; j <= radius; j++) {
						total += input.get(x+j,y+i) * dataKer[indexKer++];
					}
				}
				dataDst[indexDest++] = (short)total;
			}
		}

		// convolve along the top and bottom borders
		for ( int x = border; x < width-border; x++ ) {
			int indexDest = output.startIndex + x;

			for (int y = 0; y < border; y++, indexDest += output.stride) {
				int total = 0;
				int indexKer = 0;
				for( int i = -radius; i <= radius; i++ ) {
					for (int j = -radius; j <= radius; j++) {
						total += input.get(x+j,y+i) * dataKer[indexKer++];
					}
				}
				dataDst[indexDest] = (short)total;
			}

			indexDest = output.startIndex + (height-border) * output.stride + x;
			for (int y = height-border; y < height; y++, indexDest += output.stride) {
				int total = 0;
				int indexKer = 0;
				for( int i = -radius; i <= radius; i++ ) {
					for (int j = -radius; j <= radius; j++) {
						total += input.get(x+j,y+i) * dataKer[indexKer++];
					}
				}
				dataDst[indexDest] = (short)total;
			}
		}
	}

	public static void horizontal(Kernel1D_I32 kernel, ImageBorder_I32 input, ImageSInt32 output , int border ) {
		final int[] dataDst = output.data;
		final int[] dataKer = kernel.data;

		final int radius = kernel.getRadius();
		final int width = output.getWidth();
		final int height = output.getHeight();

		for (int y = 0; y < height; y++) {
			int indexDest = output.startIndex + y * output.stride;

			for ( int x = 0; x < border; x++ ) {
				int total = 0;
				for (int k = -radius; k <= radius; k++) {
					total += input.get(x+k,y) * dataKer[k+radius];
				}
				dataDst[indexDest++] = total;
			}

			indexDest = output.startIndex + y * output.stride + width-border;
			for ( int x = width-border; x < width; x++ ) {
				int total = 0;
				for (int k = -radius; k <= radius; k++) {
					total += input.get(x+k,y) * dataKer[k+radius];
				}
				dataDst[indexDest++] = total;
			}
		}
	}

	public static void vertical(Kernel1D_I32 kernel, ImageBorder_I32 input, ImageSInt32 output , int border ) {
		final int[] dataDst = output.data;
		final int[] dataKer = kernel.data;

		final int radius = kernel.getRadius();
		final int width = output.getWidth();
		final int height = output.getHeight();

		for ( int x = 0; x < width; x++ ) {
			int indexDest = output.startIndex + x;

			for (int y = 0; y < border; y++, indexDest += output.stride) {
				int total = 0;
				for (int k = -radius; k <= radius; k++) {
					total += input.get(x,y+k) * dataKer[k+radius];
				}
				dataDst[indexDest] = total;
			}

			indexDest = output.startIndex + (height-border) * output.stride + x;
			for (int y = height-border; y < height; y++, indexDest += output.stride) {
				int total = 0;
				for (int k = -radius; k <= radius; k++ ) {
					total += input.get(x,y+k) * dataKer[k+radius];
				}
				dataDst[indexDest] = total;
			}
		}
	}

	public static void convolve(Kernel2D_I32 kernel, ImageBorder_I32 input, ImageSInt32 output , int border ) {
		final int[] dataDst = output.data;
		final int[] dataKer = kernel.data;

		final int radius = kernel.getRadius();
		final int width = output.getWidth();
		final int height = output.getHeight();

		// convolve along the left and right borders
		for (int y = 0; y < height; y++) {
			int indexDest = output.startIndex + y * output.stride;

			for ( int x = 0; x < border; x++ ) {
				int total = 0;
				int indexKer = 0;
				for( int i = -radius; i <= radius; i++ ) {
					for (int j = -radius; j <= radius; j++) {
						total += input.get(x+j,y+i) * dataKer[indexKer++];
					}
				}
				dataDst[indexDest++] = total;
			}

			indexDest = output.startIndex + y * output.stride + width-border;
			for ( int x = width-border; x < width; x++ ) {
				int total = 0;
				int indexKer = 0;
				for( int i = -radius; i <= radius; i++ ) {
					for (int j = -radius; j <= radius; j++) {
						total += input.get(x+j,y+i) * dataKer[indexKer++];
					}
				}
				dataDst[indexDest++] = total;
			}
		}

		// convolve along the top and bottom borders
		for ( int x = border; x < width-border; x++ ) {
			int indexDest = output.startIndex + x;

			for (int y = 0; y < border; y++, indexDest += output.stride) {
				int total = 0;
				int indexKer = 0;
				for( int i = -radius; i <= radius; i++ ) {
					for (int j = -radius; j <= radius; j++) {
						total += input.get(x+j,y+i) * dataKer[indexKer++];
					}
				}
				dataDst[indexDest] = total;
			}

			indexDest = output.startIndex + (height-border) * output.stride + x;
			for (int y = height-border; y < height; y++, indexDest += output.stride) {
				int total = 0;
				int indexKer = 0;
				for( int i = -radius; i <= radius; i++ ) {
					for (int j = -radius; j <= radius; j++) {
						total += input.get(x+j,y+i) * dataKer[indexKer++];
					}
				}
				dataDst[indexDest] = total;
			}
		}
	}

}
